@@ -5,6 +5,7 @@ import itertools
 import logging
 import operator
 import random
+import re
 import sys
 import time
 from abc import ABC
@@ -20,7 +21,7 @@ from grammar_to_regex.cfg2regex import RegexConverter
 from grammar_to_regex.regex import regex_to_z3
 from orderedset import OrderedSet
 from packaging import version
-from typing import Generator, Dict, List, Set, Optional, Tuple, Union, cast, Callable
+from typing import Generator, Dict, List, Set, Optional, Tuple, Union, cast, Callable, Iterable
 
 import isla.evaluator
 import isla.helpers
@@ -35,11 +36,12 @@ from isla.helpers import delete_unreachable, dict_of_lists_to_list_of_dicts, \
     split_str_with_nonterminals, cluster_by_common_elements, is_nonterminal, canonical, lazyjoin, lazystr, is_prefix
 from isla.isla_predicates import STANDARD_STRUCTURAL_PREDICATES, STANDARD_SEMANTIC_PREDICATES, COUNT_PREDICATE
 from isla.language import VariablesCollector, split_conjunction, split_disjunction, \
-    convert_to_dnf, convert_to_nnf, ensure_unique_bound_variables, parse_isla, get_conjuncts, QuantifiedFormula
+    convert_to_dnf, convert_to_nnf, ensure_unique_bound_variables, parse_isla, get_conjuncts, QuantifiedFormula, \
+    parse_bnf
 from isla.derivation_tree import DerivationTree
 from isla.parser import EarleyParser
 from isla.three_valued_truth import ThreeValuedTruth
-from isla.type_defs import Grammar, Path
+from isla.type_defs import Grammar, Path, ImmutableList
 from isla.z3_helpers import z3_solve, z3_subst, z3_eq
 
 
@@ -166,9 +168,11 @@ class ISLaSolver:
     """
     The solver class for ISLa formulas/constraints. Main methods: `solve()` and `evaluate()`.
     """
+    UNSAT = 0
+    TIMEOUT = 1
 
     def __init__(self,
-                 grammar: Grammar,
+                 grammar: Grammar | str,
                  formula: Union[language.Formula, str],
                  structural_predicates: Set[language.StructuralPredicate] = STANDARD_STRUCTURAL_PREDICATES,
                  semantic_predicates: Set[language.SemanticPredicate] = STANDARD_SEMANTIC_PREDICATES,
@@ -183,11 +187,12 @@ class ISLaSolver:
                  global_fuzzer: bool = False,
                  predicates_unique_in_int_arg: Tuple[language.SemanticPredicate, ...] = (COUNT_PREDICATE,),
                  fuzzer_factory: Callable[[Grammar], GrammarFuzzer] = lambda grammar: GrammarCoverageFuzzer(grammar),
-                 tree_insertion_methods=DIRECT_EMBEDDING + SELF_EMBEDDING + CONTEXT_ADDITION):
+                 tree_insertion_methods=DIRECT_EMBEDDING + SELF_EMBEDDING + CONTEXT_ADDITION,
+                 activate_unsat_support: bool = False):
         """
         Constructs a new ISLaSolver object. Passing a grammar and a formula is mandatory.
 
-        :param grammar: The underlying grammar.
+        :param grammar: The underlying grammar; either, as a "Fuzzing Book" dictionary or in BNF syntax.
         :param formula: The formula to solve; either a string or a readily parsed formula.
         :param structural_predicates: Structural predicates to use when parsing a formula.
         :param semantic_predicates: Semantic predicates to use when parsing a formula.
@@ -217,6 +222,10 @@ class ISLaSolver:
         :param fuzzer_factory: Constructor of the fuzzer to use for instantiating "free" nonterminals.
         :param tree_insertion_methods: Combination of methods to use for existential quantifier elimination by
         tree insertion. Full selection: `DIRECT_EMBEDDING & SELF_EMBEDDING & CONTEXT_ADDITION`.
+        :param activate_unsat_support: Set to True if you assume that a formula might be unsatisfiable. This
+        triggers additional tests for unsatisfiability that reduce input generation performance, but might
+        ensure termination (with a negative solver result) for unsatisfiable problems for which the solver
+        could otherwise diverge.
         """
         self.logger = logging.getLogger(type(self).__name__)
 
@@ -236,21 +245,28 @@ class ISLaSolver:
             f"ISLa requires at least z3 4.8.13.0, present: {z3_version}. " \
             f"Please install a newer z3 version, e.g., using 'pip install z3-solver==4.8.14.0'."
 
-        self.grammar = grammar
-        self.graph = GrammarGraph.from_grammar(grammar)
-        self.canonical_grammar = canonical(grammar)
+        if isinstance(grammar, str):
+            self.grammar = parse_bnf(grammar)
+        else:
+            self.grammar = grammar
+
+        self.graph = GrammarGraph.from_grammar(self.grammar)
+        self.canonical_grammar = canonical(self.grammar)
         self.timeout_seconds = timeout_seconds
+        self.start_time: Optional[int] = None
         self.global_fuzzer = global_fuzzer
+        self.fuzzer = fuzzer_factory(self.grammar)
         self.fuzzer_factory = fuzzer_factory
         self.predicates_unique_in_int_arg: Set[language.SemanticPredicate] = set(predicates_unique_in_int_arg)
         self.tree_insertion_methods = tree_insertion_methods
+        self.activate_unsat_support = activate_unsat_support
 
         self.cost_computer = (
             cost_computer if cost_computer is not None
             else GrammarBasedBlackboxCostComputer(STD_COST_SETTINGS, self.graph))
 
         if isinstance(formula, str):
-            formula = parse_isla(formula, grammar, structural_predicates, semantic_predicates)
+            formula = parse_isla(formula, self.grammar, structural_predicates, semantic_predicates)
 
         self.formula = ensure_unique_bound_variables(formula)
         top_constants: Set[language.Constant] = set(
@@ -302,6 +318,8 @@ class ISLaSolver:
             heapq.heappush(self.queue, (self.compute_cost(state), state))
         self.current_level = 0
 
+        self.solutions: List[DerivationTree] = []
+
         # Debugging stuff
         self.debug = debug
         self.state_tree: Dict[SolutionState, List[SolutionState]] = {}  # is only filled if self.debug
@@ -328,22 +346,42 @@ class ISLaSolver:
         assert isinstance(inp, DerivationTree)
         return evaluate(self.formula, inp, self.grammar)
 
-    def solve(self) -> Generator[DerivationTree, None, None]:
+    def solve(self) -> Generator[DerivationTree | int, None, None]:
         """
-        Produces solutions to the constraint passed to the solver instance.
-        Take care to repeatedly call `next(...)` on the *same* generator object (don't call `solve()` repeatedly).
+        Produces solutions to the constraint passed to the solver instance, or `ISLaSolver.TIMEOUT` (if
+        a timeout occurred) or `ISLaSolver.UNSAT` if no more solutions or none at all could be found.
+        If no solution was found at all and `UNSAT` is returned, this means that the formula is unsatisfiable.
+
+        The timeout can be controlled by the `timeout_seconds` constructor parameter.
+
         :return: A generator of solutions.
         """
-        start_time = int(time.time())
-        fuzzer = self.fuzzer_factory(self.grammar)
+
+        inp = self.fuzz()
+        while isinstance(inp, DerivationTree):
+            yield inp
+            inp = self.fuzz()
+
+        yield inp
+
+    def fuzz(self) -> DerivationTree:
+        """
+        Attempts to compute a solution to the given ISLa formula. Returns that solution, if any.
+        This function can be called repeatedly to obtain more solutions until `ISLaSolver.TIMEOUT`
+        or `ISLaSolver.UNSAT` is returned. After that, only `TIMEOUT` or `UNSAT` will be returned
+        at each call.
+
+        The timeout can be controlled by the `timeout_seconds` constructor parameter.
+
+        :return: A solution for the ISLa formula passed to the `ISLaSolver`, or `ISLaSolver.TIMEOUT` (if
+        a timeout occurred) or `ISLaSolver.UNSAT` if no more solutions or none at all could be found.
+        """
+        if self.timeout_seconds is not None and self.start_time is None:
+            self.start_time = int(time.time())
 
         while self.queue:
-            if self.timeout_seconds is not None:
-                if int(time.time()) - start_time > self.timeout_seconds:
-                    return
-
             # import dill as pickle
-            # state_hash = -7249231706094547066
+            # state_hash = 9107154106757938105
             # out_file = "/tmp/saved_debug_state"
             # if hash(self.queue[0][1]) == state_hash:
             #     with open(out_file, 'wb') as debug_state_file:
@@ -351,6 +389,13 @@ class ISLaSolver:
             #     print(f"Dumping state to {out_file}")
             #     exit()
 
+            if self.timeout_seconds is not None:
+                if int(time.time()) - self.start_time > self.timeout_seconds:
+                    return ISLaSolver.TIMEOUT
+
+            if self.solutions:
+                return self.solutions.pop(0)
+
             cost: int
             state: SolutionState
             cost, state = heapq.heappop(self.queue)
@@ -371,92 +416,83 @@ class ISLaSolver:
             state = self.instantiate_structural_predicates(state)
 
             if state.constraint == sc.false():
-                continue
-
-            # Eliminate existential quantifiers over integers -> replace by constants
-            result_state = self.eliminate_existential_integer_quantifiers(state)
-            if result_state is not None:
-                yield from self.process_new_state(result_state)
-                continue
-
-            # Eliminate universal quantifiers over integers -> random choice up to max_number_free_instantiations times
-            result_states = self.instantiate_universal_integer_quantifiers(state)
-            if result_states is not None:
-                yield from self.process_new_states(result_states)
-                continue
-
-            # Match all universal formulas
-            result_states = self.match_all_universal_formulas(state)
-            if result_states is not None:
-                yield from self.process_new_states(result_states)
-                continue
-
-            # Expand if there are any not yet eliminated / matched universal quantifiers
-            if any(isinstance(conjunct, language.ForallFormula) for conjunct in get_conjuncts(state.constraint)):
+                # This state can be silently discarded.
+                pass
+            # Eliminate INTEGER EXISTENTIAL quantifiers
+            elif (result_state := self.eliminate_existential_integer_quantifiers(state)) is not None:
+                self.solutions.extend(self.process_new_state(result_state))
+            # Eliminate INTEGER UNIVERSAL quantifiers
+            elif (result_states := self.instantiate_universal_integer_quantifiers(state)) is not None:
+                self.solutions.extend(self.process_new_states(result_states))
+            # Match all UNIVERSAL FORMULAS
+            elif (result_states := self.match_all_universal_formulas(state)) is not None:
+                self.solutions.extend(self.process_new_states(result_states))
+            # EXPAND if there are any not yet eliminated / matched universal quantifiers
+            elif any(isinstance(conjunct, language.ForallFormula) for conjunct in get_conjuncts(state.constraint)):
                 expanded_states = self.expand_tree(state)
                 assert len(expanded_states) > 0, f"State {state} will never leave the queue."
-
                 self.logger.debug("Expanding state %s (%d successors)", state, len(expanded_states))
-                yield from self.process_new_states(expanded_states)
-                continue
-
-            # Eliminate all semantic formulas
-            result_states = self.eliminate_all_semantic_formulas(state)
-            if result_states is not None:
-                yield from self.process_new_states(result_states)
-                continue
 
+                self.solutions.extend(self.process_new_states(expanded_states))
+            # Eliminate all SEMANTIC FORMULAS
+            elif (result_states := self.eliminate_all_semantic_formulas(state)) is not None:
+                self.solutions.extend(self.process_new_states(result_states))
             # Eliminate all ready semantic predicate formulas
-            result_state = self.eliminate_all_ready_semantic_predicate_formulas(state)
-            if result_state is not None:
-                yield from self.process_new_state(result_state)
-                continue
-
-            # Eliminate first existential formula
-            result_states = self.eliminate_first_match_all_existential_formulas(state)
-            if result_states is not None:
-                yield from self.process_new_states(result_states)
+            elif (result_state := self.eliminate_all_ready_semantic_predicate_formulas(state)) is not None:
+                self.solutions.extend(self.process_new_state(result_state))
+            # Eliminate first EXISTENTIAL FORMULA
+            elif (result_states := self.eliminate_and_match_first_existential_formula(state)) is not None:
+                self.solutions.extend(self.process_new_states(result_states))
                 # Also add some expansions of the original state, to create a larger
                 # solution stream (otherwise, it might be possible that only a small
                 # finite number of solutions are generated for existential formulas).
-                yield from self.process_new_states(self.expand_tree(state, limit=2, only_universal=False))
-                continue
-
-            # semantic predicate formulas can remain if they bind lazily. In that case, we can choose a random
-            # instantiation and let the predicate "fix" the resulting tree.
-            assert (state.constraint == sc.true() or
-                    all(isinstance(conjunct, language.SemanticPredicateFormula)
-                        or (isinstance(conjunct, language.NegatedFormula) and
-                            isinstance(conjunct.args[0], language.SemanticPredicateFormula))
-                        for conjunct in get_conjuncts(state.constraint))), \
-                f"Constraint is not true and contains formulas " \
-                f"other than semantic predicate formulas: {state.constraint}"
-            assert (state.constraint == sc.true() or
-                    all(not pred_formula.binds_tree(leaf)
-                        for pred_formula in get_conjuncts(state.constraint)
-                        if isinstance(pred_formula, language.SemanticPredicateFormula)
-                        for _, leaf in state.tree.open_leaves()) or
-                    all(not cast(language.SemanticPredicateFormula, pred_formula.args[0]).binds_tree(leaf)
-                        for pred_formula in get_conjuncts(state.constraint)
-                        if isinstance(pred_formula, language.NegatedFormula) and
-                        isinstance(pred_formula.args[0], language.SemanticPredicateFormula))
-                    for _, leaf in state.tree.open_leaves()
-                    ), \
-                f"Constraint is not true and contains semantic predicate formulas which bind open leaves in the tree: " \
-                f"{state.constraint}, leaves: {', '.join(list(map(str, [leaf for _, leaf in state.tree.open_leaves()])))}"
-
-            if not self.global_fuzzer:
-                fuzzer = self.fuzzer_factory(self.grammar)
-            if state.constraint == sc.true():
-                for _ in range(self.max_number_free_instantiations):
-                    result = state.tree
-                    for path, leaf in state.tree.open_leaves():
-                        leaf_inst = fuzzer.expand_tree(DerivationTree(leaf.value, None))
-                        result = result.replace_path(path, leaf_inst)
-                    yield from self.process_new_states([SolutionState(state.constraint, result)])
+                self.solutions.extend(self.process_new_states(self.expand_tree(state, limit=2, only_universal=False)))
             else:
-                # noinspection PyTypeChecker
-                yield from self.process_new_states(self.expand_state(state, fuzzer))
+                # SEMANTIC PREDICATE FORMULAS can remain if they bind lazily. In that case, we can choose a random
+                # instantiation and let the predicate "fix" the resulting tree.
+                assert (state.constraint == sc.true() or
+                        all(isinstance(conjunct, language.SemanticPredicateFormula)
+                            or (isinstance(conjunct, language.NegatedFormula) and
+                                isinstance(conjunct.args[0], language.SemanticPredicateFormula))
+                            for conjunct in get_conjuncts(state.constraint))), \
+                    f"Constraint is not true and contains formulas " \
+                    f"other than semantic predicate formulas: {state.constraint}"
+                assert (state.constraint == sc.true() or
+                        all(not pred_formula.binds_tree(leaf)
+                            for pred_formula in get_conjuncts(state.constraint)
+                            if isinstance(pred_formula, language.SemanticPredicateFormula)
+                            for _, leaf in state.tree.open_leaves()) or
+                        all(not cast(language.SemanticPredicateFormula, pred_formula.args[0]).binds_tree(leaf)
+                            for pred_formula in get_conjuncts(state.constraint)
+                            if isinstance(pred_formula, language.NegatedFormula) and
+                            isinstance(pred_formula.args[0], language.SemanticPredicateFormula))
+                        for _, leaf in state.tree.open_leaves()
+                        ), \
+                    f"Constraint is not true and contains semantic predicate formulas binding open tree leaves: " \
+                    f"{state.constraint}, leaves: " + \
+                    str({', '.join(list(map(str, [leaf for _, leaf in state.tree.open_leaves()])))})
+
+                if self.global_fuzzer:
+                    fuzzer = self.fuzzer
+                else:
+                    fuzzer = self.fuzzer_factory(self.grammar)
+
+                if state.constraint == sc.true():
+                    for _ in range(self.max_number_free_instantiations):
+                        result = state.tree
+                        for path, leaf in state.tree.open_leaves():
+                            leaf_inst = fuzzer.expand_tree(DerivationTree(leaf.value, None))
+                            result = result.replace_path(path, leaf_inst)
+
+                        self.solutions.extend(self.process_new_state(SolutionState(state.constraint, result)))
+                else:
+                    # noinspection PyTypeChecker
+                    self.solutions.extend(self.process_new_states(self.expand_state(state, fuzzer)))
+
+        if self.solutions:
+            return self.solutions.pop(0)
+        else:
+            return ISLaSolver.UNSAT
 
     def expand_state(self, state: SolutionState, fuzzer: GrammarCoverageFuzzer) -> List[SolutionState]:
         result: List[SolutionState] = []
@@ -563,8 +599,7 @@ class ISLaSolver:
 
             try:
                 solver_result = self.solve_quantifier_free_formula(
-                    [smt_formula], max_instantiations=max_instantiations
-                )
+                    (smt_formula,), max_instantiations=max_instantiations)
 
                 solutions: Dict[language.Constant, Set[int]] = {
                     c: {int(solution[cast(language.Constant, c)].value)
@@ -592,7 +627,7 @@ class ISLaSolver:
                     return {
                         c for c in candidates
                         if self.solve_quantifier_free_formula(
-                            [cast(language.SMTFormula, smt_formula.substitute_variables({constant: c}))],
+                            (cast(language.SMTFormula, smt_formula.substitute_variables({constant: c})),),
                             max_instantiations=1)}
 
         instantiations: List[Dict[language.Constant | DerivationTree,
@@ -754,7 +789,19 @@ class ISLaSolver:
             universal_int_formula)
         return []
 
-    def eliminate_all_semantic_formulas(self, state: SolutionState) -> Optional[List[SolutionState]]:
+    def eliminate_all_semantic_formulas(
+            self,
+            state: SolutionState,
+            max_instantiations: Optional[int] = None) -> Optional[List[SolutionState]]:
+        """
+        Eliminates all SMT-LIB formulas that appear in `state`'s constraint as conjunctive elements.
+        If, e.g., an SMT-LIB formula occurs as a disjunction, no solution is computed.
+
+        :param state: The state in which to solve all SMT-LIB formulas.
+        :param max_instantiations: The number of solutions the SMT solver should be asked for.
+        :return: The discovered solutions.
+        """
+
         conjuncts = split_conjunction(state.constraint)
         semantic_formulas = [conjunct for conjunct in conjuncts
                              if isinstance(conjunct, language.SMTFormula)
@@ -774,7 +821,8 @@ class ISLaSolver:
 
         return self.eliminate_semantic_formula(
             prefix_conjunction,
-            SolutionState(new_disjunct, state.tree))
+            SolutionState(new_disjunct, state.tree),
+            max_instantiations)
 
     def eliminate_all_ready_semantic_predicate_formulas(self, state: SolutionState) -> Optional[SolutionState]:
         semantic_predicate_formulas: List[language.NegatedFormula | language.SemanticPredicateFormula] = [
@@ -840,12 +888,9 @@ class ISLaSolver:
 
         return result if changed else None
 
-    def eliminate_first_match_all_existential_formulas(self, state: SolutionState) -> Optional[List[SolutionState]]:
-        # We produce up to three output states: One where all existential formulas that can be matched
-        # are matched, one where only the first existential formula is matched, and one where only the
-        # first existential formula is eliminated by tree insertion. The reason is that for re-inserted
-        # existential formulas, eager matching is the best strategy; in other cases, however, it is not
-        # clear whether matching or insertion are the best strategies.
+    def eliminate_and_match_first_existential_formula(self, state: SolutionState) -> Optional[List[SolutionState]]:
+        # We produce up to two groups of output states: One where the first existential formula, if it can be matched,
+        # is matched, and one where the first existential formula is eliminated by tree insertion.
         existential_formulas = [
             conjunct for conjunct in split_conjunction(state.constraint)
             if isinstance(conjunct, language.ExistsFormula)]
@@ -885,6 +930,10 @@ class ISLaSolver:
         # 2. Match first existential formula.
         first_matched = OrderedSet(self.match_existential_formula(existential_formulas[0], state))
 
+        # Tree insertion can be deactivated by setting `self.tree_insertion_methods` to 0.
+        if not self.tree_insertion_methods:
+            return list(first_matched)
+
         if first_matched:
             self.logger.debug(
                 "Matched first existential formulas, result: [%s]",
@@ -957,6 +1006,9 @@ class ISLaSolver:
                 or (not only_universal and isinstance(formula, language.QuantifiedFormula))
             )}
 
+        if not nonterminal_expansions:
+            return []
+
         possible_expansions: List[Dict[Path, List[DerivationTree]]] = []
         if not limit:
             possible_expansions = dict_of_lists_to_list_of_dicts(nonterminal_expansions)
@@ -1178,7 +1230,10 @@ class ISLaSolver:
         return result
 
     def eliminate_semantic_formula(
-            self, semantic_formula: language.Formula, state: SolutionState) -> List[SolutionState]:
+            self,
+            semantic_formula: language.Formula,
+            state: SolutionState,
+            max_instantiations: Optional[int] = None) -> Optional[List[SolutionState]]:
         """
         Solves a semantic formula and, for each solution, substitutes the solution for the respective
         constant in each assignment of the state. Also instantiates all "free" constants in the given
@@ -1188,6 +1243,7 @@ class ISLaSolver:
         :param semantic_formula: The semantic (i.e., only containing logical connectors and SMT Formulas)
         formula to solve.
         :param state: The original solution state.
+        :param max_instantiations: The maximum number of solutions to ask the SMT solver for.
         :return: A list of instantiated SolutionStates.
         """
 
@@ -1224,7 +1280,8 @@ class ISLaSolver:
             formula_clusters.append(remaining_clusters)
 
         all_solutions: List[List[Dict[Union[language.Constant, DerivationTree], DerivationTree]]] = [
-            self.solve_quantifier_free_formula(cluster) for cluster in formula_clusters
+            self.solve_quantifier_free_formula(tuple(cluster), max_instantiations)
+            for cluster in formula_clusters
         ]
 
         # These solutions are all independent, such that we can combine each solution with all others.
@@ -1265,22 +1322,36 @@ class ISLaSolver:
 
         return results
 
+    @lru_cache(100)
     def solve_quantifier_free_formula(
             self,
-            smt_formulas: List[language.SMTFormula],
+            smt_formulas: ImmutableList[language.SMTFormula],
             max_instantiations: Optional[int] = None) -> \
             List[Dict[language.Constant | DerivationTree, DerivationTree]]:
-        # If any SMT formula refers to subtrees in the instantiations of other SMT formulas,
+        """
+        Attempts to solve the given SMT-LIB formulas by calling Z3.
+
+        Note that this function does not unify variables pointing to the same derivation trees.
+        E.g., a solution may be returned for the formula `var_1 = "a" and var_2 = "b"`, though
+        `var_1` and `var_2` point to the same `<var>` tree as defined by their substitutions maps.
+        Unification is performed in `eliminate_all_semantic_formulas`.
+
+        :param smt_formulas: The SMT-LIB formulas to solve.
+        :param max_instantiations: The maximum number of instantiations to produce.
+        :return: A (possibly empty) list of solutions.
+        """
+
+        # If any SMT formula refers to *sub*trees in the instantiations of other SMT formulas,
         # we have to instantiate those first.
-        def get_conflicts(smt_formulas):
+        def get_conflicts(smt_formulas: Iterable[language.SMTFormula]):
             return [
-                f for fidx, f in enumerate(smt_formulas)
+                formula for formula_idx, formula in enumerate(smt_formulas)
                 if any(
-                    ot != t and ot.find_node(t) is not None
-                    for t in f.substitutions.values()
-                    for ofidx, of in enumerate(smt_formulas)
-                    if fidx != ofidx
-                    for ot in of.substitutions.values())]
+                    other_subst_tree != subst_tree and other_subst_tree.find_node(subst_tree) is not None
+                    for subst_tree in formula.substitutions.values()
+                    for other_formula_idx, other_formula in enumerate(smt_formulas)
+                    if formula_idx != other_formula_idx
+                    for other_subst_tree in other_formula.substitutions.values())]
 
         conflicts = get_conflicts(smt_formulas)
 
@@ -1410,6 +1481,13 @@ class ISLaSolver:
         new_states = [self.remove_nonmatching_universal_quantifiers(new_state) for new_state in new_states]
         new_states = [self.remove_infeasible_universal_quantifiers(new_state) for new_state in new_states]
 
+        if self.activate_unsat_support:
+            for new_state in list(new_states):
+                # Remove states with unsatisfiable SMT-LIB formulas.
+                if (any(isinstance(f, language.SMTFormula) for f in split_conjunction(new_state.constraint)) and
+                        not self.eliminate_all_semantic_formulas(new_state, max_instantiations=1)):
+                    new_states.remove(new_state)
+
         assert all(
             state.tree.find_node(tree) is not None
             for state in new_states
@@ -1549,11 +1627,11 @@ class ISLaSolver:
             if not isinstance(universal_formula, language.ForallFormula):
                 continue
 
-            if not (any(self.quantified_formula_might_match(universal_formula, leaf_path, universal_formula.in_variable)
-                        for leaf_path, leaf_node in universal_formula.in_variable.open_leaves())
-                    or [match for match in
-                        isla.evaluator.matches_for_quantified_formula(universal_formula, self.grammar)
-                        if not universal_formula.is_already_matched(match[universal_formula.bound_variable][1])]):
+            if (all(not self.quantified_formula_might_match(universal_formula, leaf_path, universal_formula.in_variable)
+                    for leaf_path, leaf_node in universal_formula.in_variable.open_leaves()) and
+                    not [match for match in
+                         isla.evaluator.matches_for_quantified_formula(universal_formula, self.grammar)
+                         if not universal_formula.is_already_matched(match[universal_formula.bound_variable][1])]):
                 result = SolutionState(
                     language.replace_formula(result.constraint, universal_formula, sc.true()),
                     result.tree)
@@ -1791,7 +1869,23 @@ def quantified_formula_might_match(
         tree: DerivationTree,
         grammar: Grammar,
         reachable: Callable[[str, str], bool]) -> bool:
+    """
+    This function returns `True` if in order to match the `qfd_formula`, the given path
+    has to be expanded. It will return `False` if expanding that path makes no difference,
+    either because there is a match already or because the expansion will not contribute
+    to a future match. The purpose is to check, e.g., if a nonterminal shall be expanded
+     or "freely" instantiated; or to check if a quantifier can be removed, because it won't
+      match any expansion (in addition of not already matchine!).
+
+    :param qfd_formula: The quantified formula for which to check if the given tree path might become a match.
+    :param path_to_nonterminal: The path in the tree to check.
+    :param tree: The context tree.
+    :param grammar: The reference grammar.
+    :param reachable: A reachability function in the grammar.
+    :return: `True` iff some expansion of this path matches the formula, and the formula does not already match.
+    """
     node = tree.get_subtree(path_to_nonterminal)
+    assert not node.children, 'quantified_formula_might_match only works for leaf nodes'
 
     if qfd_formula.in_variable.find_node(node) is None:
         return False
@@ -1802,17 +1896,20 @@ def quantified_formula_might_match(
                    for path, _ in node.paths() if path)
 
     qfd_nonterminal = qfd_formula.bound_variable.n_type
-    if qfd_nonterminal == node.value or reachable(node.value, qfd_nonterminal):
+
+    if qfd_nonterminal == node.value:
+        return qfd_formula.bind_expression is not None
+
+    if qfd_nonterminal != node.value and reachable(node.value, qfd_nonterminal):
         return True
 
     if qfd_formula.bind_expression is None:
-        return False
+        return False  # The formula matches a (parent) element or not, but it's no future match.
 
-    # Is there an extension of some tree `node` is a subtree of, such that the
-    # bind expression tree is a prefix tree of that extension, and extending
-    # the nonterminal might approach a match?
-    # Note that the prefix tree must not *already* be a match.
+    # This leaf won't reach the "root node" of the match tree for `qfd_formula`.
+    assert qfd_nonterminal != node.value and not reachable(node.value, qfd_nonterminal)
 
+    # Return true if extending this leaf makes the quantifier match some parent.
     maybe_prefix_tree: DerivationTree
     for maybe_prefix_tree, var_map in qfd_formula.bind_expression.to_tree_prefix(
             qfd_formula.bound_variable.n_type, grammar):
